package com.opensymphony.oscache.plugins.memcached;

import com.opensymphony.oscache.base.persistence.PersistenceListener;
import com.opensymphony.oscache.base.persistence.CachePersistenceException;
import com.opensymphony.oscache.base.Config;
import com.danga.MemCached.SockIOPool;
import com.danga.MemCached.MemCachedClient;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


import java.util.*;

/**
 * Persist the cache data to memcached servers.
 *
 * @version $Id: MemcachedListener.java,v 1.6 2008/12/17 10:24:35 fabrizio Exp $
 * @author <a href="mailto:fzellini@gmail.com">Fabrizio Zellini</a>
 *
 *  Copyright 2008 Fabrizio Zellini
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */


public class MemcachedListener implements PersistenceListener {

    private final Log log = LogFactory.getLog(this.getClass());

    public static final int BLOCKING=-99;

    public static final String INSTANCE_NAME="cache.memcached.instance_name";
    public static final String DEFAULT_INSTANCE_NAME="memos";

    public static final String SERVERS="cache.memcached.servers";
    public static final String DEFAULT_SERVERS="localhost:11211";

    public static final String INITCONN="cache.memcached.initconn";
    public static final int DEFAULT_INITCONN=100;


    public static final String MINCONN="cache.memcached.minconn";
    public static final int DEFAULT_MINCONN=100;

    public static final String MAXCONN="cache.memcached.maxconn";
    public static final int DEFAULT_MAXCONN=500;

    public static final String MAINTSLEEP="cache.memcached.maintsleep";
    public static final int DEFAULT_MAINTSLEEP=30;

    public static final String SOCK_CONNECT_TO="cache.memcached.sock_connect_to";
    public static final int DEFAULT_SOCK_CONNECT_TO=BLOCKING;

    public static final String SOCK_READ_TO="cache.memcached.sock_read_to";
    public static final int DEFAULT_SOCK_READ_TO=BLOCKING;


    public static final String NAGLE="cache.memcached.nagle";
    public static final boolean DEFAULT_NAGLE=false;

    public static final String COMPRESS="cache.memcached.compress";
    public static final boolean DEFAULT_COMPRESS=false;

    public static final String KEY_PREFIX="cache.memcached.keyprefix";
    public static final String DEFAULT_KEYPREFIX="_oscache_";

    public static final String KEY_ALGO="cache.memcached.keyalgo";
    public static final int DEFAULT_KEYALGO=0;

    private String instanceName;
    private String servers;
    private int initConns;
    private int minConns;
    private int maxConns;
    private int maintSleep;
    private int sockConnectTo;
    private int sockReadTo;
    private boolean nagle;
    private boolean compress=false;
    private String keyPrefix;

    private int keyAlgo;

    private SockIOPool pool;
    private MemCachedClient mclient;


    private String makeKey (String key){
        if (key==null) return null;
        String thekey;
        switch (keyAlgo){
            case 1:
                MD5 md = new MD5();
                byte[] bytes = key.getBytes();
                md.update(bytes);

                byte out[] = new byte[16];
                md.md5final(out);
                thekey = keyPrefix+MD5.dumpBytes(out);
                break;
            case 0:
            default:
                thekey = keyPrefix+key;
                break;
        }
        if (log.isDebugEnabled()){
            log.debug("returning key : ["+thekey+"] for inputkey ["+key+"]");
        }
        return thekey;
    }

    protected String getStringProperty (Config config,String propname,String defaultvalue){
        String rt=defaultvalue;
        String s= config.getProperty(propname);
        if (s!=null) rt = s;
        return rt;
    }
    protected int getIntProperty (Config config,String propname, int defaultvalue){
        int rt = defaultvalue;
        String value = getStringProperty(config,propname,null);
        try {
            if (value!=null)
                rt = Integer.parseInt(value);
        } catch (Exception e){
            if (log.isWarnEnabled()){
                log.warn("Error parsing ["+propname+"], value is ["+value+"]",e);
            }
        }
        return rt;
    }
    protected boolean getBooleanProperty (Config config,String propname, boolean defaultvalue){
        boolean rt = defaultvalue;
        String value = getStringProperty(config,propname,null);
        if (value!=null){
            if ("false".equalsIgnoreCase(value)){
                rt = false;
            } else
            if ("true".equalsIgnoreCase(value)){
                rt = true;
            } else {
                if (log.isWarnEnabled()){
                    log.warn("Error parsing ["+propname+"], value is ["+value+"], expected boolean");
                }
            }
        }
        return rt;
    }

    /**
     * split a String
     * @param s  the string to split
     * @param delim e.g. ";"
     * @return array of tokens
     */
    protected String[] splitString (String s, String delim){
        StringTokenizer tokenizer = new StringTokenizer(s,delim);
        int tokens = tokenizer.countTokens();
        String[] rt = null;
        if (tokens>0){
            rt = new String[tokens];
            int i=0;
            while (tokenizer.hasMoreElements()){
                rt[i++] = tokenizer.nextToken();
            }
        }
        return rt;
    }

    private String getGroupKey (String group){
        return keyPrefix+"__group__"+group;
        //return "__group__"+group;
    }

    /**
     * Verify if an object is currently stored in the persistent cache.
     *
     * @param key The cache key of the object to check.
     */
    public boolean isStored(String key) throws CachePersistenceException{

        return mclient.keyExists(makeKey(key));

    }

    /**
     * Verify if a group is currently stored in the persistent cache.
     *
     * @param groupName The name of the group to check.
     */
    public boolean isGroupStored(String groupName) throws CachePersistenceException{
        return isStored(getGroupKey(groupName));
    }

    /**
     * Clear the entire persistent cache (including the root)
     */
    public void clear() throws CachePersistenceException{
        mclient.flushAll();
    }

    /**
    * Initialises this <tt>MemcachedListener</tt> using the supplied
    * configuration.
    *
    * @param config The OSCache configuration
    */
    public PersistenceListener configure(Config config){

        instanceName = getStringProperty(config,INSTANCE_NAME,DEFAULT_INSTANCE_NAME);
        if (log.isDebugEnabled()){
            log.debug("instanceName : ["+instanceName+"]");
        }

        servers = getStringProperty(config,SERVERS,DEFAULT_SERVERS);
        if (log.isDebugEnabled()){
            log.debug("servers : ["+servers+"]");
        }

        keyPrefix = getStringProperty(config,KEY_PREFIX,DEFAULT_KEYPREFIX);
        if (log.isDebugEnabled()){
            log.debug("keyprefix : ["+keyPrefix+"]");
        }

        initConns = getIntProperty(config,INITCONN,DEFAULT_INITCONN);
        if (log.isDebugEnabled()){
            log.debug("initConns : ["+initConns+"]");
        }

        minConns = getIntProperty(config,MINCONN,DEFAULT_MINCONN);
        if (log.isDebugEnabled()){
            log.debug("minConns  : ["+minConns+"]");
        }


        maxConns = getIntProperty(config,MAXCONN,DEFAULT_MAXCONN);
        if (log.isDebugEnabled()){
            log.debug("maxConns  : ["+maxConns+"]");
        }

        maintSleep = getIntProperty(config,MAINTSLEEP,DEFAULT_MAINTSLEEP);
        if (log.isDebugEnabled()){
            log.debug("maintSleep  : ["+maintSleep+"]");
        }

        sockConnectTo = getIntProperty(config,SOCK_CONNECT_TO,DEFAULT_SOCK_CONNECT_TO);
        if (log.isDebugEnabled()){
            log.debug("sock_connect_timeout : ["+sockConnectTo+"]");
        }
        
        sockReadTo = getIntProperty(config,SOCK_READ_TO,DEFAULT_SOCK_READ_TO);
        if (log.isDebugEnabled()){
            log.debug("sock_read_timeout : ["+sockReadTo+"]");
        }

        nagle = getBooleanProperty(config,NAGLE,DEFAULT_NAGLE);
        if (log.isDebugEnabled()){
            log.debug("nagle  : ["+nagle+"]");
        }

        compress = getBooleanProperty(config,COMPRESS,DEFAULT_COMPRESS);
        if (log.isDebugEnabled()){
            log.debug("compress  : ["+compress+"]");
        }

        keyAlgo = getIntProperty(config,KEY_ALGO,0);
        if (log.isDebugEnabled()){
            log.debug("keyAlgo  : ["+keyAlgo+"]");
        }

        pool = SockIOPool.getInstance(this.instanceName);
        pool.setServers(splitString(servers,";"));
        pool.setInitConn(initConns);
        pool.setMinConn(minConns);
        pool.setMaxConn(maxConns);
        pool.setMaintSleep(maintSleep);
        pool.setNagle(nagle);
        
        if (sockReadTo!=BLOCKING){
            pool.setSocketTO(sockReadTo);
        }
        if (sockConnectTo!=BLOCKING){
            pool.setSocketConnectTO(sockConnectTo);
        }

        pool.initialize();

        mclient = new MemCachedClient(instanceName);
        //mclient.setPoolName(instanceName);  // really need ?
        mclient.setCompressEnable(compress);


        mclient.setCompressEnable( false );

        return this;
    }

    /**
     * Removes an object from the persistent cache
     */
    public void remove(String key) throws CachePersistenceException{
        mclient.delete(makeKey(key));
    }

    /**
     * Removes a group from the persistent cache.
     *
     * @param groupName The name of the cache group to remove.
     */
    public void removeGroup(String groupName) throws CachePersistenceException {
        remove(getGroupKey(groupName));
    }

    /**
     * Retrieves an object from the persistent cache.
     *
     * @param key The unique cache key that maps to the object
     * being retrieved.
     * @return The object, or <code>null</code> if no object was found
     * matching the supplied key.
     */
    public Object retrieve(String key) throws CachePersistenceException{
        return mclient.get(makeKey(key));
    }

    /**
     * Stores an object in the persistent cache.
     *
     * @param key The key to uniquely identify this object.
     * @param obj The object to persist. Most implementations
     * of this interface will require this object implements
     * <code>Serializable</code>.
     */
    public void store(String key, Object obj) throws CachePersistenceException{
        mclient.add(makeKey(key),obj);
        //mclient.add(key,obj);
    }

    /**
     * Stores a group in the persistent cache.
     *
     * @param groupName The name of the group to persist.
     * @param group A set containing the keys of all the <code>CacheEntry</code>
     * objects that belong to this group.
     */
    public void storeGroup(String groupName, Set group) throws CachePersistenceException{
        store (getGroupKey(groupName),group);
    }

    /**
     * Retrieves a group from the persistent cache.
     *
     * @param groupName The name of the group to retrieve.
     * @return The returned set should contain the keys
     * of all the <code>CacheEntry</code> objects that belong
     * to this group.
     */
    public Set retrieveGroup(String groupName) throws CachePersistenceException{
        return (Set) retrieve(getGroupKey(groupName));
    }

}
